"""
If you use this code, please cite the first SynthSeg paper:
https://github.com/BBillot/lab2im/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import numpy as np
import numpy.random as npr

# project imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import edit_volumes
from lamareg.ext.lab2im.lab2im_model import lab2im_model


class ImageGenerator:

    def __init__(
        self,
        labels_dir,
        generation_labels=None,
        output_labels=None,
        batchsize=1,
        n_channels=1,
        target_res=None,
        output_shape=None,
        output_div_by_n=None,
        generation_classes=None,
        prior_distributions="uniform",
        prior_means=None,
        prior_stds=None,
        use_specific_stats_for_channel=False,
        blur_range=1.15,
    ):
        """
        This class is wrapper around the lab2im_model model. It contains the GPU model that generates images from labels
        maps, and a python generator that supplies the input data for this model.
        To generate pairs of image/labels you can just call the method generate_image() on an object of this class.

        :param labels_dir: path of folder with all input label maps, or to a single label map.

        # IMPORTANT !!!
        # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
        # these values refer to the RAS axes.

        # label maps-related parameters
        :param generation_labels: (optional) list of all possible label values in the input label maps.
        Default is None, where the label values are directly gotten from the provided label maps.
        If not None, can be a sequence or a 1d numpy array, or the path to a 1d numpy array.
        :param output_labels: (optional) list of the same length as generation_labels to indicate which values to use in
        the label maps returned by this function, i.e. all occurrences of generation_labels[i] in the input label maps
        will be converted to output_labels[i] in the returned label maps. Examples:
        Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
        Set output_labels[i]=generation_labels[i] to keep the value generation_labels[i] in the returned maps.
        Can be a list or a 1d numpy array. By default output_labels is equal to generation_labels.

        # output-related parameters
        :param batchsize: (optional) numbers of images to generate per mini-batch. Default is 1.
        :param n_channels: (optional) number of channels to be synthetised. Default is 1.
        :param target_res: (optional) target resolution of the generated images and corresponding label maps.
        If None, the outputs will have the same resolution as the input label maps.
        Can be a number (isotropic resolution), a sequence, a 1d numpy array, or the path to a 1d numpy array.
        :param output_shape: (optional) shape of the output image, obtained by randomly cropping the generated image.
        Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
        :param output_div_by_n: (optional) forces the output shape to be divisible by this value. It overwrites
        output_shape if necessary. Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or
        the path to a 1d numpy array.

        # GMM-sampling parameters
        :param generation_classes: (optional) Indices regrouping generation labels into classes of same intensity
        distribution. Regrouped labels will thus share the same Gaussian when sampling a new image. Can be a sequence, a
        1d numpy array, or the path to a 1d numpy array.
        It should have the same length as generation_labels, and contain values between 0 and K-1, where K is the total
        number of classes. Default is all labels have different classes (K=len(generation_labels)).
        :param prior_distributions: (optional) type of distribution from which we sample the GMM parameters.
        Can either be 'uniform', or 'normal'. Default is 'uniform'.
        :param prior_means: (optional) hyperparameters controlling the prior distributions of the GMM means. Because
        these prior distributions are uniform or normal, they require by 2 hyperparameters. Thus prior_means can be:
        1) a sequence of length 2, directly defining the two hyperparameters: [min, max] if prior_distributions is
        uniform, [mean, std] if the distribution is normal. The GMM means of are independently sampled at each
        mini_batch from the same distribution.
        2) an array of shape (2, K), where K is the number of classes (K=len(generation_labels) if generation_classes is
        not given). The mean of the Gaussian distribution associated to class k in [0, ...K-1] is sampled at each
        mini-batch from U(prior_means[0,k], prior_means[1,k]) if prior_distributions is uniform, and from
        N(prior_means[0,k], prior_means[1,k]) if prior_distributions is normal.
        3) an array of shape (2*n_mod, K), where each block of two rows is associated to hyperparameters derived
        from different modalities. In this case, if use_specific_stats_for_channel is False, we first randomly select a
        modality from the n_mod possibilities, and we sample the GMM means like in 2).
        If use_specific_stats_for_channel is True, each block of two rows correspond to a different channel
        (n_mod=n_channels), thus we select the corresponding block to each channel rather than randomly drawing it.
        4) the path to such a numpy array.
        Default is None, which corresponds to prior_means = [25, 225].
        :param prior_stds: (optional) same as prior_means but for the standard deviations of the GMM.
        Default is None, which corresponds to prior_stds = [5, 25].
        :param use_specific_stats_for_channel: (optional) whether the i-th block of two rows in the prior arrays must be
        only used to generate the i-th channel. If True, n_mod should be equal to n_channels. Default is False.

        # blurring parameters
        :param blur_range: (optional) Randomise the standard deviation of the blurring kernels, (whether data_res is
        given or not). At each mini_batch, the standard deviation of the blurring kernels are multiplied by a c
        coefficient sampled from a uniform distribution with bounds [1/blur_range, blur_range].
        If None, no randomisation. Default is 1.15.
        """

        # prepare data files
        self.labels_paths = utils.list_images_in_folder(labels_dir)

        # generation parameters
        self.labels_shape, self.aff, self.n_dims, _, self.header, self.atlas_res = (
            utils.get_volume_info(self.labels_paths[0], aff_ref=np.eye(4))
        )
        self.n_channels = n_channels
        if generation_labels is not None:
            self.generation_labels = utils.load_array_if_path(generation_labels)
        else:
            self.generation_labels, _ = utils.get_list_labels(labels_dir=labels_dir)
        if output_labels is not None:
            self.output_labels = utils.load_array_if_path(output_labels)
        else:
            self.output_labels = self.generation_labels
        self.target_res = utils.load_array_if_path(target_res)
        self.batchsize = batchsize
        # preliminary operations
        self.output_shape = utils.load_array_if_path(output_shape)
        self.output_div_by_n = output_div_by_n
        # GMM parameters
        self.prior_distributions = prior_distributions
        if generation_classes is not None:
            self.generation_classes = utils.load_array_if_path(generation_classes)
            assert (
                self.generation_classes.shape == self.generation_labels.shape
            ), "if provided, generation labels should have the same shape as generation_labels"
            unique_classes = np.unique(self.generation_classes)
            assert np.array_equal(
                unique_classes, np.arange(np.max(unique_classes) + 1)
            ), "generation_classes should a linear range between 0 and its maximum value."
        else:
            self.generation_classes = np.arange(self.generation_labels.shape[0])
        self.prior_means = utils.load_array_if_path(prior_means)
        self.prior_stds = utils.load_array_if_path(prior_stds)
        self.use_specific_stats_for_channel = use_specific_stats_for_channel

        # blurring parameters
        self.blur_range = blur_range

        # build transformation model
        self.labels_to_image_model, self.model_output_shape = self._build_lab2im_model()

        # build generator for model inputs
        self.model_inputs_generator = self._build_model_inputs(
            len(self.generation_labels)
        )

        # build brain generator
        self.image_generator = self._build_image_generator()

    def _build_lab2im_model(self):
        # build_model
        lab_to_im_model = lab2im_model(
            labels_shape=self.labels_shape,
            n_channels=self.n_channels,
            generation_labels=self.generation_labels,
            output_labels=self.output_labels,
            atlas_res=self.atlas_res,
            target_res=self.target_res,
            output_shape=self.output_shape,
            output_div_by_n=self.output_div_by_n,
            blur_range=self.blur_range,
        )
        out_shape = lab_to_im_model.output[0].get_shape().as_list()[1:]
        return lab_to_im_model, out_shape

    def _build_image_generator(self):
        while True:
            model_inputs = next(self.model_inputs_generator)
            [image, labels] = self.labels_to_image_model.predict(model_inputs)
            yield image, labels

    def generate_image(self):
        """call this method when an object of this class has been instantiated to generate new brains"""
        (image, labels) = next(self.image_generator)
        # put back images in native space
        list_images = list()
        list_labels = list()
        for i in range(self.batchsize):
            list_images.append(
                edit_volumes.align_volume_to_ref(
                    image[i], np.eye(4), aff_ref=self.aff, n_dims=self.n_dims
                )
            )
            list_labels.append(
                edit_volumes.align_volume_to_ref(
                    labels[i], np.eye(4), aff_ref=self.aff, n_dims=self.n_dims
                )
            )
        image = np.stack(list_images, axis=0)
        labels = np.stack(list_labels, axis=0)
        return np.squeeze(image), np.squeeze(labels)

    def _build_model_inputs(self, n_labels):

        # get label info
        _, _, n_dims, _, _, _ = utils.get_volume_info(self.labels_paths[0])

        # Generate!
        while True:

            # randomly pick as many images as batchsize
            indices = npr.randint(len(self.labels_paths), size=self.batchsize)

            # initialise input lists
            list_label_maps = []
            list_means = []
            list_stds = []

            for idx in indices:

                # load label in identity space, and add them to inputs
                y = utils.load_volume(
                    self.labels_paths[idx], dtype="int", aff_ref=np.eye(4)
                )
                list_label_maps.append(utils.add_axis(y, axis=[0, -1]))

                # add means and standard deviations to inputs
                means = np.empty((1, n_labels, 0))
                stds = np.empty((1, n_labels, 0))
                for channel in range(self.n_channels):

                    # retrieve channel specific stats if necessary
                    if isinstance(self.prior_means, np.ndarray):
                        if (
                            self.prior_means.shape[0] > 2
                        ) & self.use_specific_stats_for_channel:
                            if self.prior_means.shape[0] / 2 != self.n_channels:
                                raise ValueError(
                                    "the number of blocks in prior_means does not match n_channels. This "
                                    "message is printed because use_specific_stats_for_channel is True."
                                )
                            tmp_prior_means = self.prior_means[
                                2 * channel : 2 * channel + 2, :
                            ]
                        else:
                            tmp_prior_means = self.prior_means
                    else:
                        tmp_prior_means = self.prior_means
                    if isinstance(self.prior_stds, np.ndarray):
                        if (
                            self.prior_stds.shape[0] > 2
                        ) & self.use_specific_stats_for_channel:
                            if self.prior_stds.shape[0] / 2 != self.n_channels:
                                raise ValueError(
                                    "the number of blocks in prior_stds does not match n_channels. This "
                                    "message is printed because use_specific_stats_for_channel is True."
                                )
                            tmp_prior_stds = self.prior_stds[
                                2 * channel : 2 * channel + 2, :
                            ]
                        else:
                            tmp_prior_stds = self.prior_stds
                    else:
                        tmp_prior_stds = self.prior_stds

                    # draw means and std devs from priors
                    tmp_classes_means = utils.draw_value_from_distribution(
                        tmp_prior_means,
                        n_labels,
                        self.prior_distributions,
                        125.0,
                        100.0,
                        positive_only=True,
                    )
                    tmp_classes_stds = utils.draw_value_from_distribution(
                        tmp_prior_stds,
                        n_labels,
                        self.prior_distributions,
                        15.0,
                        10.0,
                        positive_only=True,
                    )
                    tmp_means = utils.add_axis(
                        tmp_classes_means[self.generation_classes], axis=[0, -1]
                    )
                    tmp_stds = utils.add_axis(
                        tmp_classes_stds[self.generation_classes], axis=[0, -1]
                    )
                    means = np.concatenate([means, tmp_means], axis=-1)
                    stds = np.concatenate([stds, tmp_stds], axis=-1)
                list_means.append(means)
                list_stds.append(stds)

            # build list of inputs of augmentation model
            list_inputs = [list_label_maps, list_means, list_stds]
            if (
                self.batchsize > 1
            ):  # concatenate individual input types if batchsize > 1
                list_inputs = [np.concatenate(item, 0) for item in list_inputs]
            else:
                list_inputs = [item[0] for item in list_inputs]

            yield list_inputs



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import numpy as np

# project imports
from lamareg.SynthSeg.model_inputs import build_model_inputs
from lamareg.SynthSeg.labels_to_image_model import labels_to_image_model

# third-party imports
from lamareg.ext.lab2im import utils, edit_volumes


class BrainGenerator:

    def __init__(
        self,
        labels_dir,
        generation_labels=None,
        n_neutral_labels=None,
        output_labels=None,
        subjects_prob=None,
        batchsize=1,
        n_channels=1,
        target_res=None,
        output_shape=None,
        output_div_by_n=None,
        prior_distributions="uniform",
        generation_classes=None,
        prior_means=None,
        prior_stds=None,
        use_specific_stats_for_channel=False,
        mix_prior_and_random=False,
        flipping=True,
        scaling_bounds=0.2,
        rotation_bounds=15,
        shearing_bounds=0.012,
        translation_bounds=False,
        nonlin_std=4.0,
        nonlin_scale=0.04,
        randomise_res=True,
        max_res_iso=4.0,
        max_res_aniso=8.0,
        data_res=None,
        thickness=None,
        bias_field_std=0.7,
        bias_scale=0.025,
        return_gradients=False,
    ):
        """
        This class is wrapper around the labels_to_image_model model. It contains the GPU model that generates images
        from labels maps, and a python generator that supplies the input data for this model.
        To generate pairs of image/labels you can just call the method generate_image() on an object of this class.

        :param labels_dir: path of folder with all input label maps, or to a single label map.

        # IMPORTANT !!!
        # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
        # these values refer to the RAS axes.

        # label maps-related parameters
        :param generation_labels: (optional) list of all possible label values in the input label maps.
        Default is None, where the label values are directly gotten from the provided label maps.
        If not None, can be a sequence or a 1d numpy array, or the path to a 1d numpy array.
        If flipping is true (i.e. right/left flipping is enabled), generation_labels should be organised as follows:
        background label first, then non-sided labels (e.g. CSF, brainstem, etc.), then all the structures of the same
        hemisphere (can be left or right), and finally all the corresponding contralateral structures in the same order.
        :param n_neutral_labels: (optional) number of non-sided generation labels. This is important only if you use
        flipping augmentation. Default is total number of label values.
        :param output_labels: (optional) list of the same length as generation_labels to indicate which values to use in
        the label maps returned by this function, i.e. all occurrences of generation_labels[i] in the input label maps
        will be converted to output_labels[i] in the returned label maps. Examples:
        Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
        Set output_labels[i]=generation_labels[i] to keep the value generation_labels[i] in the returned maps.
        Can be a list or a 1d numpy array. By default output_labels is equal to generation_labels.
        :param subjects_prob: (optional) relative order of importance (doesn't have to be probabilistic), with which to
        pick the provided label maps at each minibatch. Can be a sequence, a 1D numpy array, or the path to such an
        array, and it must be as long as path_label_maps. By default, all label maps are chosen with the same importance

        # output-related parameters
        :param batchsize: (optional) numbers of images to generate per mini-batch. Default is 1.
        :param n_channels: (optional) number of channels to be synthesised. Default is 1.
        :param target_res: (optional) target resolution of the generated images and corresponding label maps.
        If None, the outputs will have the same resolution as the input label maps.
        Can be a number (isotropic resolution), a sequence, a 1d numpy array, or the path to a 1d numpy array.
        :param output_shape: (optional) shape of the output image, obtained by randomly cropping the generated image.
        Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
        Default is None, where no cropping is performed.
        :param output_div_by_n: (optional) forces the output shape to be divisible by this value. It overwrites
        output_shape if necessary. Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or
        the path to a 1d numpy array.

        # GMM-sampling parameters
        :param generation_classes: (optional) Indices regrouping generation labels into classes of same intensity
        distribution. Regrouped labels will thus share the same Gaussian when sampling a new image. Can be a sequence, a
        1d numpy array, or the path to a 1d numpy array. It should have the same length as generation_labels, and
        contain values between 0 and K-1, where K is the total number of classes.
        Default is all labels have different classes (K=len(generation_labels)).
        :param prior_distributions: (optional) type of distribution from which we sample the GMM parameters.
        Can either be 'uniform', or 'normal'. Default is 'uniform'.
        :param prior_means: (optional) hyperparameters controlling the prior distributions of the GMM means. Because
        these prior distributions are uniform or normal, they require by 2 hyperparameters. Thus prior_means can be:
        1) a sequence of length 2, directly defining the two hyperparameters: [min, max] if prior_distributions is
        uniform, [mean, std] if the distribution is normal. The GMM means of are independently sampled at each
        mini_batch from the same distribution.
        2) an array of shape (2, K), where K is the number of classes (K=len(generation_labels) if generation_classes is
        not given). The mean of the Gaussian distribution associated to class k in [0, ...K-1] is sampled at each
        mini-batch from U(prior_means[0,k], prior_means[1,k]) if prior_distributions is uniform, and from
        N(prior_means[0,k], prior_means[1,k]) if prior_distributions is normal.
        3) an array of shape (2*n_mod, K), where each block of two rows is associated to hyperparameters derived
        from different modalities. In this case, if use_specific_stats_for_channel is False, we first randomly select a
        modality from the n_mod possibilities, and we sample the GMM means like in 2).
        If use_specific_stats_for_channel is True, each block of two rows correspond to a different channel
        (n_mod=n_channels), thus we select the corresponding block to each channel rather than randomly drawing it.
        4) the path to such a numpy array.
        Default is None, which corresponds to prior_means = [25, 225].
        :param prior_stds: (optional) same as prior_means but for the standard deviations of the GMM.
        Default is None, which corresponds to prior_stds = [5, 25].
        :param use_specific_stats_for_channel: (optional) whether the i-th block of two rows in the prior arrays must be
        only used to generate the i-th channel. If True, n_mod should be equal to n_channels. Default is False.
        :param mix_prior_and_random: (optional) if prior_means is not None, enables to reset the priors to their default
        values for half of these cases, and thus generate images of random contrast.

        # spatial deformation parameters
        :param flipping: (optional) whether to introduce right/left random flipping. Default is True.
        :param scaling_bounds: (optional) range of the random sampling to apply at each mini-batch. The scaling factor
        for each dimension is sampled from a uniform distribution of predefined bounds. Can either be:
        1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
        [1-scaling_bounds, 1+scaling_bounds] for each dimension.
        2) a sequence, in which case the scaling factor is sampled from the uniform distribution of bounds
        (1-scaling_bounds[i], 1+scaling_bounds[i]) for the i-th dimension.
        3) a numpy array of shape (2, n_dims), in which case the scaling factor is sampled from the uniform distribution
         of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
        4) False, in which case scaling is completely turned off.
        Default is scaling_bounds = 0.2 (case 1)
        :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for cases 1
        and 2, the bounds are centred on 0 rather than 1, i.e. [0+rotation_bounds[i], 0-rotation_bounds[i]].
        Default is rotation_bounds = 15.
        :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
        :param translation_bounds: (optional) same as scaling bounds. Default is translation_bounds = False, but we
        encourage using it when cropping is deactivated (i.e. when output_shape=None in BrainGenerator).
        :param nonlin_std: (optional) Maximum value for the standard deviation of the normal distribution from which we
        sample the first tensor for synthesising the deformation field. Set to 0 if you wish to completely turn the
        elastic deformation off.
        :param nonlin_scale: (optional) if nonlin_std is strictly positive, factor between the shapes of the
        input label maps and the shape of the input non-linear tensor.

        # blurring/resampling parameters
        :param randomise_res: (optional) whether to mimic images that would have been 1) acquired at low resolution, and
        2) resampled to high resolution. The low resolution is uniformly resampled at each minibatch from [1mm, 9mm].
        In that process, the images generated by sampling the GMM are:
        1) blurred at the sampled LR, 2) downsampled at LR, and 3) resampled at target_resolution.
        :param max_res_iso: (optional) If randomise_res is True, this enables to control the upper bound of the uniform
        distribution from which we sample the random resolution U(min_res, max_res_iso), where min_res is the resolution
        of the input label maps. Must be a number, and default is 4. Set to None to deactivate it, but if randomise_res
        is True, at least one of max_res_iso or max_res_aniso must be given.
        :param max_res_aniso: If randomise_res is True, this enables to downsample the input volumes to a random LR
        in only 1 (random) direction. This is done by randomly selecting a direction i in range [0, n_dims-1], and
        sampling a value in the corresponding uniform distribution U(min_res[i], max_res_aniso[i]), where min_res is the
        resolution of the input label maps. Can be a number, a sequence, or a 1d numpy array. Set to None to deactivate
        it, but if randomise_res is True, at least one of max_res_iso or max_res_aniso must be given.
        :param data_res: (optional) specific acquisition resolution to mimic, as opposed to random resolution sampled
        when randomise_res is True. This triggers a blurring which mimics the acquisition resolution, but downsampling
        is optional (see param downsample). Default for data_res is None, where images are slightly blurred.
        If the generated images are uni-modal, data_res can be a number (isotropic acquisition resolution), a sequence,
        a 1d numpy array, or the path to a 1d numpy array. In the multi-modal case, it should be given as a numpy array
        (or a path) of size (n_mod, n_dims), where each row is the acquisition resolution of the corresponding channel.
        :param thickness: (optional) if data_res is provided, we can further specify the slice thickness of the low
        resolution images to mimic. Must be provided in the same format as data_res. Default thickness = data_res.

        # bias field parameters
        :param bias_field_std: (optional) If strictly positive, this triggers the corruption of synthesised images with
        a bias field. It is obtained by sampling a first small tensor from a normal distribution, resizing it to full
        size, and rescaling it to positive values by taking the voxel-wise exponential. bias_field_std designates the
        std dev of the normal distribution from which we sample the first tensor. Set to 0 to deactivate bias field.
        :param bias_scale: (optional) If bias_field_std is strictly positive, this designates the ratio between
        the size of the input label maps and the size of the first sampled tensor for synthesising the bias field.

        :param return_gradients: (optional) whether to return the synthetic image or the magnitude of its spatial
        gradient (computed with Sobel kernels).
        """

        # prepare data files
        self.labels_paths = utils.list_images_in_folder(labels_dir)
        if subjects_prob is not None:
            self.subjects_prob = np.array(
                utils.reformat_to_list(subjects_prob, load_as_numpy=True),
                dtype="float32",
            )
            assert len(self.subjects_prob) == len(self.labels_paths), (
                "subjects_prob should have the same length as labels_path, "
                "had {} and {}".format(len(self.subjects_prob), len(self.labels_paths))
            )
        else:
            self.subjects_prob = None

        # generation parameters
        self.labels_shape, self.aff, self.n_dims, _, self.header, self.atlas_res = (
            utils.get_volume_info(self.labels_paths[0], aff_ref=np.eye(4))
        )
        self.n_channels = n_channels
        if generation_labels is not None:
            self.generation_labels = utils.load_array_if_path(generation_labels)
        else:
            self.generation_labels, _ = utils.get_list_labels(labels_dir=labels_dir)
        if output_labels is not None:
            self.output_labels = utils.load_array_if_path(output_labels)
        else:
            self.output_labels = self.generation_labels
        if n_neutral_labels is not None:
            self.n_neutral_labels = n_neutral_labels
        else:
            self.n_neutral_labels = self.generation_labels.shape[0]
        self.target_res = utils.load_array_if_path(target_res)
        self.batchsize = batchsize
        # preliminary operations
        self.flipping = flipping
        self.output_shape = utils.load_array_if_path(output_shape)
        self.output_div_by_n = output_div_by_n
        # GMM parameters
        self.prior_distributions = prior_distributions
        if generation_classes is not None:
            self.generation_classes = utils.load_array_if_path(generation_classes)
            assert (
                self.generation_classes.shape == self.generation_labels.shape
            ), "if provided, generation_classes should have the same shape as generation_labels"
            unique_classes = np.unique(self.generation_classes)
            assert np.array_equal(
                unique_classes, np.arange(np.max(unique_classes) + 1)
            ), "generation_classes should a linear range between 0 and its maximum value."
        else:
            self.generation_classes = np.arange(self.generation_labels.shape[0])
        self.prior_means = utils.load_array_if_path(prior_means)
        self.prior_stds = utils.load_array_if_path(prior_stds)
        self.use_specific_stats_for_channel = use_specific_stats_for_channel
        # linear transformation parameters
        self.scaling_bounds = utils.load_array_if_path(scaling_bounds)
        self.rotation_bounds = utils.load_array_if_path(rotation_bounds)
        self.shearing_bounds = utils.load_array_if_path(shearing_bounds)
        self.translation_bounds = utils.load_array_if_path(translation_bounds)
        # elastic transformation parameters
        self.nonlin_std = nonlin_std
        self.nonlin_scale = nonlin_scale
        # blurring parameters
        self.randomise_res = randomise_res
        self.max_res_iso = max_res_iso
        self.max_res_aniso = max_res_aniso
        self.data_res = utils.load_array_if_path(data_res)
        assert not (
            self.randomise_res & (self.data_res is not None)
        ), "randomise_res and data_res cannot be provided at the same time"
        self.thickness = utils.load_array_if_path(thickness)
        # bias field parameters
        self.bias_field_std = bias_field_std
        self.bias_scale = bias_scale
        self.return_gradients = return_gradients

        # build transformation model
        self.labels_to_image_model, self.model_output_shape = (
            self._build_labels_to_image_model()
        )

        # build generator for model inputs
        self.model_inputs_generator = self._build_model_inputs_generator(
            mix_prior_and_random
        )

        # build brain generator
        self.brain_generator = self._build_brain_generator()

    def _build_labels_to_image_model(self):
        # build_model
        lab_to_im_model = labels_to_image_model(
            labels_shape=self.labels_shape,
            n_channels=self.n_channels,
            generation_labels=self.generation_labels,
            output_labels=self.output_labels,
            n_neutral_labels=self.n_neutral_labels,
            atlas_res=self.atlas_res,
            target_res=self.target_res,
            output_shape=self.output_shape,
            output_div_by_n=self.output_div_by_n,
            flipping=self.flipping,
            aff=np.eye(4),
            scaling_bounds=self.scaling_bounds,
            rotation_bounds=self.rotation_bounds,
            shearing_bounds=self.shearing_bounds,
            translation_bounds=self.translation_bounds,
            nonlin_std=self.nonlin_std,
            nonlin_scale=self.nonlin_scale,
            randomise_res=self.randomise_res,
            max_res_iso=self.max_res_iso,
            max_res_aniso=self.max_res_aniso,
            data_res=self.data_res,
            thickness=self.thickness,
            bias_field_std=self.bias_field_std,
            bias_scale=self.bias_scale,
            return_gradients=self.return_gradients,
        )
        out_shape = lab_to_im_model.output[0].get_shape().as_list()[1:]
        return lab_to_im_model, out_shape

    def _build_model_inputs_generator(self, mix_prior_and_random):
        # build model's inputs generator
        model_inputs_generator = build_model_inputs(
            path_label_maps=self.labels_paths,
            n_labels=len(self.generation_labels),
            batchsize=self.batchsize,
            n_channels=self.n_channels,
            subjects_prob=self.subjects_prob,
            generation_classes=self.generation_classes,
            prior_means=self.prior_means,
            prior_stds=self.prior_stds,
            prior_distributions=self.prior_distributions,
            use_specific_stats_for_channel=self.use_specific_stats_for_channel,
            mix_prior_and_random=mix_prior_and_random,
        )
        return model_inputs_generator

    def _build_brain_generator(self):
        while True:
            model_inputs = next(self.model_inputs_generator)
            [image, labels] = self.labels_to_image_model.predict(model_inputs)
            yield image, labels

    def generate_brain(self):
        """call this method when an object of this class has been instantiated to generate new brains"""
        (image, labels) = next(self.brain_generator)
        # put back images in native space
        list_images = list()
        list_labels = list()
        for i in range(self.batchsize):
            list_images.append(
                edit_volumes.align_volume_to_ref(
                    image[i], np.eye(4), aff_ref=self.aff, n_dims=self.n_dims
                )
            )
            list_labels.append(
                edit_volumes.align_volume_to_ref(
                    labels[i], np.eye(4), aff_ref=self.aff, n_dims=self.n_dims
                )
            )
        image = np.squeeze(np.stack(list_images, axis=0))
        labels = np.squeeze(np.stack(list_labels, axis=0))
        return image, labels



"""
tensorflow/keras utilities for the neuron project

If you use this code, please cite 
Dalca AV, Guttag J, Sabuncu MR
Anatomical Priors in Convolutional Networks for Unsupervised Biomedical Segmentation, 
CVPR 2018

or for the transformation/interpolation related functions:

Unsupervised Learning for Fast Probabilistic Diffeomorphic Registration
Adrian V. Dalca, Guha Balakrishnan, John Guttag, Mert R. Sabuncu
MICCAI 2018.

Contact: adalca [at] csail [dot] mit [dot] edu
License: GPLv3
"""

import itertools
import numpy as np
import tensorflow as tf
import keras.backend as K


def interpn(vol, loc, interp_method='linear'):
    """
    N-D gridded interpolation in tensorflow

    vol can have more dimensions than loc[i], in which case loc[i] acts as a slice 
    for the first dimensions

    Parameters:
        vol: volume with size vol_shape or [*vol_shape, nb_features]
        loc: an N-long list of N-D Tensors (the interpolation locations) for the new grid
            each tensor has to have the same size (but not nec. same size as vol)
            or a tensor of size [*new_vol_shape, D]
        interp_method: interpolation type 'linear' (default) or 'nearest'

    Returns:
        new interpolated volume of the same size as the entries in loc
    """

    if isinstance(loc, (list, tuple)):
        loc = tf.stack(loc, -1)
    nb_dims = loc.shape[-1]

    if len(vol.shape) not in [nb_dims, nb_dims + 1]:
        raise Exception("Number of loc Tensors %d does not match volume dimension %d"
                        % (nb_dims, len(vol.shape[:-1])))

    if nb_dims > len(vol.shape):
        raise Exception("Loc dimension %d does not match volume dimension %d"
                        % (nb_dims, len(vol.shape)))

    if len(vol.shape) == nb_dims:
        vol = K.expand_dims(vol, -1)

    # flatten and float location Tensors
    loc = tf.cast(loc, 'float32')

    if isinstance(vol.shape, tf.TensorShape):
        volshape = vol.shape.as_list()
    else:
        volshape = vol.shape

    # interpolate
    if interp_method == 'linear':
        loc0 = tf.floor(loc)

        # clip values
        max_loc = [d - 1 for d in vol.get_shape().as_list()]
        clipped_loc = [tf.clip_by_value(loc[..., d], 0, max_loc[d]) for d in range(nb_dims)]
        loc0lst = [tf.clip_by_value(loc0[..., d], 0, max_loc[d]) for d in range(nb_dims)]

        # get other end of point cube
        loc1 = [tf.clip_by_value(loc0lst[d] + 1, 0, max_loc[d]) for d in range(nb_dims)]
        locs = [[tf.cast(f, 'int32') for f in loc0lst], [tf.cast(f, 'int32') for f in loc1]]

        # compute the difference between the upper value and the original value
        # differences are basically 1 - (pt - floor(pt))
        #   because: floor(pt) + 1 - pt = 1 + (floor(pt) - pt) = 1 - (pt - floor(pt))
        diff_loc1 = [loc1[d] - clipped_loc[d] for d in range(nb_dims)]
        diff_loc0 = [1 - d for d in diff_loc1]
        weights_loc = [diff_loc1, diff_loc0]  # note reverse ordering since weights are inverse of diff.

        # go through all the cube corners, indexed by a ND binary vector 
        # e.g. [0, 0] means this "first" corner in a 2-D "cube"
        cube_pts = list(itertools.product([0, 1], repeat=nb_dims))
        interp_vol = 0

        for c in cube_pts:
            # get nd values
            # note re: indices above volumes via https://github.com/tensorflow/tensorflow/issues/15091
            #   It works on GPU because we do not perform index validation checking on GPU -- it's too
            #   expensive. Instead we fill the output with zero for the corresponding value. The CPU
            #   version caught the bad index and returned the appropriate error.
            subs = [locs[c[d]][d] for d in range(nb_dims)]

            idx = sub2ind(vol.shape[:-1], subs)
            vol_val = tf.gather(tf.reshape(vol, [-1, volshape[-1]]), idx)

            # get the weight of this cube_pt based on the distance
            # if c[d] is 0 --> want weight = 1 - (pt - floor[pt]) = diff_loc1
            # if c[d] is 1 --> want weight = pt - floor[pt] = diff_loc0
            wts_lst = [weights_loc[c[d]][d] for d in range(nb_dims)]
            wt = prod_n(wts_lst)
            wt = K.expand_dims(wt, -1)

            # compute final weighted value for each cube corner
            interp_vol += wt * vol_val

    else:
        assert interp_method == 'nearest'
        roundloc = tf.cast(tf.round(loc), 'int32')

        # clip values
        max_loc = [tf.cast(d - 1, 'int32') for d in vol.shape]
        roundloc = [tf.clip_by_value(roundloc[..., d], 0, max_loc[d]) for d in range(nb_dims)]

        # get values
        idx = sub2ind(vol.shape[:-1], roundloc)
        interp_vol = tf.gather(tf.reshape(vol, [-1, vol.shape[-1]]), idx)

    return interp_vol


def resize(vol, zoom_factor, new_shape, interp_method='linear'):
    """
    if zoom_factor is a list, it will determine the ndims, in which case vol has to be of length ndims or ndims + 1

    if zoom_factor is an integer, then vol must be of length ndims + 1

    new_shape should be a list of length ndims

    """

    if isinstance(zoom_factor, (list, tuple)):
        ndims = len(zoom_factor)
        vol_shape = vol.shape[:ndims]
        assert len(vol_shape) in (ndims, ndims + 1), \
            "zoom_factor length %d does not match ndims %d" % (len(vol_shape), ndims)
    else:
        vol_shape = vol.shape[:-1]
        ndims = len(vol_shape)
        zoom_factor = [zoom_factor] * ndims

    # get grid for new shape
    grid = volshape_to_ndgrid(new_shape)
    grid = [tf.cast(f, 'float32') for f in grid]
    offset = [grid[f] / zoom_factor[f] - grid[f] for f in range(ndims)]
    offset = tf.stack(offset, ndims)

    # transform
    return transform(vol, offset, interp_method)


zoom = resize


def affine_to_shift(affine_matrix, volshape, shift_center=True, indexing='ij'):
    """
    transform an affine matrix to a dense location shift tensor in tensorflow

    Algorithm:
        - get grid and shift grid to be centered at the center of the image (optionally)
        - apply affine matrix to each index.
        - subtract grid

    Parameters:
        affine_matrix: ND+1 x ND+1 or ND x ND+1 matrix (Tensor)
        volshape: 1xN Nd Tensor of the size of the volume.
        shift_center (optional)
        indexing

    Returns:
        shift field (Tensor) of size *volshape x N
    """

    if isinstance(volshape, tf.TensorShape):
        volshape = volshape.as_list()

    if affine_matrix.dtype != 'float32':
        affine_matrix = tf.cast(affine_matrix, 'float32')

    nb_dims = len(volshape)

    if len(affine_matrix.shape) == 1:
        if len(affine_matrix) != (nb_dims * (nb_dims + 1)):
            raise ValueError('transform is supposed a vector of len ndims * (ndims + 1).'
                             'Got len %d' % len(affine_matrix))

        affine_matrix = tf.reshape(affine_matrix, [nb_dims, nb_dims + 1])

    if not (affine_matrix.shape[0] in [nb_dims, nb_dims + 1] and affine_matrix.shape[1] == (nb_dims + 1)):
        raise Exception('Affine matrix shape should match'
                        '%d+1 x %d+1 or ' % (nb_dims, nb_dims) +
                        '%d x %d+1.' % (nb_dims, nb_dims) +
                        'Got: ' + str(volshape))

    # list of volume ndgrid
    # N-long list, each entry of shape volshape
    mesh = volshape_to_meshgrid(volshape, indexing=indexing)
    mesh = [tf.cast(f, 'float32') for f in mesh]

    if shift_center:
        mesh = [mesh[f] - (volshape[f] - 1) / 2 for f in range(len(volshape))]

    # add an all-ones entry and transform into a large matrix
    flat_mesh = [flatten(f) for f in mesh]
    flat_mesh.append(tf.ones(flat_mesh[0].shape, dtype='float32'))
    mesh_matrix = tf.transpose(tf.stack(flat_mesh, axis=1))  # 4 x nb_voxels

    # compute locations
    loc_matrix = tf.matmul(affine_matrix, mesh_matrix)  # N+1 x nb_voxels
    loc_matrix = tf.transpose(loc_matrix[:nb_dims, :])  # nb_voxels x N
    loc = tf.reshape(loc_matrix, list(volshape) + [nb_dims])  # *volshape x N

    # get shifts and return
    return loc - tf.stack(mesh, axis=nb_dims)


def combine_non_linear_and_aff_to_shift(transform_list, volshape, shift_center=True, indexing='ij'):
    """
    transform an affine matrix to a dense location shift tensor in tensorflow

    Algorithm:
        - get grid and shift grid to be centered at the center of the image (optionally)
        - apply affine matrix to each index.
        - subtract grid

    Parameters:
        transform_list: list of non-linear tensor (size of volshape) and affine ND+1 x ND+1 or ND x ND+1 tensor
        volshape: 1xN Nd Tensor of the size of the volume.
        shift_center (optional)
        indexing

    Returns:
        shift field (Tensor) of size *volshape x N
    """

    if isinstance(volshape, tf.TensorShape):
        volshape = volshape.as_list()

    # convert transforms to floats
    for i in range(len(transform_list)):
        if transform_list[i].dtype != 'float32':
            transform_list[i] = tf.cast(transform_list[i], 'float32')

    nb_dims = len(volshape)

    # transform affine to matrix if given as vector
    if len(transform_list[1].shape) == 1:
        if len(transform_list[1]) != (nb_dims * (nb_dims + 1)):
            raise ValueError('transform is supposed a vector of len ndims * (ndims + 1).'
                             'Got len %d' % len(transform_list[1]))

        transform_list[1] = tf.reshape(transform_list[1], [nb_dims, nb_dims + 1])

    if not (transform_list[1].shape[0] in [nb_dims, nb_dims + 1] and transform_list[1].shape[1] == (nb_dims + 1)):
        raise Exception('Affine matrix shape should match'
                        '%d+1 x %d+1 or ' % (nb_dims, nb_dims) +
                        '%d x %d+1.' % (nb_dims, nb_dims) +
                        'Got: ' + str(volshape))

    # list of volume ndgrid
    # N-long list, each entry of shape volshape
    mesh = volshape_to_meshgrid(volshape, indexing=indexing)
    mesh = [tf.cast(f, 'float32') for f in mesh]

    if shift_center:
        mesh = [mesh[f] - (volshape[f] - 1) / 2 for f in range(len(volshape))]

    # add an all-ones entry and transform into a large matrix
    # non_linear_mesh = tf.unstack(transform_list[0], axis=3)
    non_linear_mesh = tf.unstack(transform_list[0], axis=-1)
    flat_mesh = [flatten(mesh[i]+non_linear_mesh[i]) for i in range(len(mesh))]
    flat_mesh.append(tf.ones(flat_mesh[0].shape, dtype='float32'))
    mesh_matrix = tf.transpose(tf.stack(flat_mesh, axis=1))  # N+1 x nb_voxels

    # compute locations
    loc_matrix = tf.matmul(transform_list[1], mesh_matrix)  # N+1 x nb_voxels
    loc_matrix = tf.transpose(loc_matrix[:nb_dims, :])  # nb_voxels x N
    loc = tf.reshape(loc_matrix, list(volshape) + [nb_dims])  # *volshape x N

    # get shifts and return
    return loc - tf.stack(mesh, axis=nb_dims)


def transform(vol, loc_shift, interp_method='linear', indexing='ij'):
    """
    transform interpolation N-D volumes (features) given shifts at each location in tensorflow

    Essentially interpolates volume vol at locations determined by loc_shift. 
    This is a spatial transform in the sense that at location [x] we now have the data from, 
    [x + shift] so we've moved data.

    Parameters:
        vol: volume with size vol_shape or [*vol_shape, nb_features]
        loc_shift: shift volume [*new_vol_shape, N]
        interp_method (default:'linear'): 'linear', 'nearest'
        indexing (default: 'ij'): 'ij' (matrix) or 'xy' (cartesian).
            In general, prefer to leave this 'ij'
    
    Return:
        new interpolated volumes in the same size as loc_shift[0]
    """

    # parse shapes
    if isinstance(loc_shift.shape, tf.TensorShape):
        volshape = loc_shift.shape[:-1].as_list()
    else:
        volshape = loc_shift.shape[:-1]
    nb_dims = len(volshape)

    # location should be meshed and delta
    mesh = volshape_to_meshgrid(volshape, indexing=indexing)  # volume mesh
    loc = [tf.cast(mesh[d], 'float32') + loc_shift[..., d] for d in range(nb_dims)]

    # test single
    return interpn(vol, loc, interp_method=interp_method)


def integrate_vec(vec, time_dep=False, method='ss', **kwargs):
    """
    Integrate (stationary of time-dependent) vector field (N-D Tensor) in tensorflow
    
    Aside from directly using tensorflow's numerical integration odeint(), also implements 
    "scaling and squaring", and quadrature. Note that the diff. equation given to odeint
    is the one used in quadrature.   

    Parameters:
        vec: the Tensor field to integrate. 
            If vol_size is the size of the intrinsic volume, and vol_ndim = len(vol_size),
            then vector shape (vec_shape) should be 
            [vol_size, vol_ndim] (if stationary)
            [vol_size, vol_ndim, nb_time_steps] (if time dependent)
        time_dep: bool whether vector is time dependent
        method: 'scaling_and_squaring' or 'ss' or 'quadrature'
        
        if using 'scaling_and_squaring': currently only supports integrating to time point 1.
            nb_steps int number of steps. Note that this means the vec field gets broken own to 2**nb_steps.
            so nb_steps of 0 means integral = vec.

    Returns:
        int_vec: integral of vector field with same shape as the input
    """

    if method not in ['ss', 'scaling_and_squaring', 'ode', 'quadrature']:
        raise ValueError("method has to be 'scaling_and_squaring' or 'ode'. found: %s" % method)

    if method in ['ss', 'scaling_and_squaring']:
        nb_steps = kwargs['nb_steps']
        assert nb_steps >= 0, 'nb_steps should be >= 0, found: %d' % nb_steps

        if time_dep:
            svec = K.permute_dimensions(vec, [-1, *range(0, vec.shape[-1] - 1)])
            assert 2 ** nb_steps == svec.shape[0], "2**nb_steps and vector shape don't match"

            svec = svec / (2 ** nb_steps)
            for _ in range(nb_steps):
                svec = svec[0::2] + tf.map_fn(transform, svec[1::2, :], svec[0::2, :])

            disp = svec[0, :]

        else:
            vec = vec / (2 ** nb_steps)
            for _ in range(nb_steps):
                vec += transform(vec, vec)
            disp = vec

    else:  # method == 'quadrature':
        nb_steps = kwargs['nb_steps']
        assert nb_steps >= 1, 'nb_steps should be >= 1, found: %d' % nb_steps

        vec = vec / nb_steps

        if time_dep:
            disp = vec[..., 0]
            for si in range(nb_steps - 1):
                disp += transform(vec[..., si + 1], disp)
        else:
            disp = vec
            for _ in range(nb_steps - 1):
                disp += transform(vec, disp)

    return disp


def volshape_to_ndgrid(volshape, **kwargs):
    """
    compute Tensor ndgrid from a volume size

    Parameters:
        volshape: the volume size

    Returns:
        A list of Tensors

    See Also:
        ndgrid
    """

    isint = [float(d).is_integer() for d in volshape]
    if not all(isint):
        raise ValueError("volshape needs to be a list of integers")

    linvec = [tf.range(0, d) for d in volshape]
    return ndgrid(*linvec, **kwargs)


def volshape_to_meshgrid(volshape, **kwargs):
    """
    compute Tensor meshgrid from a volume size

    Parameters:
        volshape: the volume size

    Returns:
        A list of Tensors

    See Also:
        tf.meshgrid, meshgrid, ndgrid, volshape_to_ndgrid
    """

    isint = [float(d).is_integer() for d in volshape]
    if not all(isint):
        raise ValueError("volshape needs to be a list of integers")

    linvec = [tf.range(0, d) for d in volshape]
    return meshgrid(*linvec, **kwargs)


def ndgrid(*args, **kwargs):
    """
    broadcast Tensors on an N-D grid with ij indexing
    uses meshgrid with ij indexing

    Parameters:
        *args: Tensors with rank 1
        **args: "name" (optional)

    Returns:
        A list of Tensors
    
    """
    return meshgrid(*args, indexing='ij', **kwargs)


def meshgrid(*args, **kwargs):
    """
    
    meshgrid code that builds on (copies) tensorflow's meshgrid but dramatically
    improves runtime by changing the last step to tiling instead of multiplication.
    https://github.com/tensorflow/tensorflow/blob/c19e29306ce1777456b2dbb3a14f511edf7883a8/tensorflow/python/ops/array_ops.py#L1921
    
    Broadcasts parameters for evaluation on an N-D grid.
    Given N one-dimensional coordinate arrays `*args`, returns a list `outputs`
    of N-D coordinate arrays for evaluating expressions on an N-D grid.
    Notes:
    `meshgrid` supports cartesian ('xy') and matrix ('ij') indexing conventions.
    When the `indexing` argument is set to 'xy' (the default), the broadcasting
    instructions for the first two dimensions are swapped.
    Examples:
    Calling `X, Y = meshgrid(x, y)` with the tensors
    ```python
    x = [1, 2, 3]
    y = [4, 5, 6]
    X, Y = meshgrid(x, y)
    # X = [[1, 2, 3],
    #      [1, 2, 3],
    #      [1, 2, 3]]
    # Y = [[4, 4, 4],
    #      [5, 5, 5],
    #      [6, 6, 6]]
    ```
    Args:
    *args: `Tensor`s with rank 1.
    **kwargs:
      - indexing: Either 'xy' or 'ij' (optional, default: 'xy').
      - name: A name for the operation (optional).
    Returns:
    outputs: A list of N `Tensor`s with rank N.
    Raises:
    TypeError: When no keyword arguments (kwargs) are passed.
    ValueError: When indexing keyword argument is not one of `xy` or `ij`.
    """

    indexing = kwargs.pop("indexing", "xy")
    if kwargs:
        key = list(kwargs.keys())[0]
        raise TypeError("'{}' is an invalid keyword argument "
                        "for this function".format(key))

    if indexing not in ("xy", "ij"):
        raise ValueError("indexing parameter must be either 'xy' or 'ij'")

    # with ops.name_scope(name, "meshgrid", args) as name:
    ndim = len(args)
    s0 = (1,) * ndim

    # Prepare reshape by inserting dimensions with size 1 where needed
    output = []
    for i, x in enumerate(args):
        output.append(tf.reshape(tf.stack(x), (s0[:i] + (-1,) + s0[i + 1::])))
    # Create parameters for broadcasting each tensor to the full size
    shapes = [tf.size(x) for x in args]
    sz = [x.get_shape().as_list()[0] for x in args]

    # output_dtype = tf.convert_to_tensor(args[0]).dtype.base_dtype
    if indexing == "xy" and ndim > 1:
        output[0] = tf.reshape(output[0], (1, -1) + (1,) * (ndim - 2))
        output[1] = tf.reshape(output[1], (-1, 1) + (1,) * (ndim - 2))
        shapes[0], shapes[1] = shapes[1], shapes[0]
        sz[0], sz[1] = sz[1], sz[0]

    for i in range(len(output)):
        stack_sz = [*sz[:i], 1, *sz[(i + 1):]]
        if indexing == 'xy' and ndim > 1 and i < 2:
            stack_sz[0], stack_sz[1] = stack_sz[1], stack_sz[0]
        output[i] = tf.tile(output[i], tf.stack(stack_sz))
    return output


def flatten(v):
    """flatten Tensor v"""

    return tf.reshape(v, [-1])


def prod_n(lst):
    prod = lst[0]
    for p in lst[1:]:
        prod *= p
    return prod


def sub2ind(siz, subs):
    """assumes column-order major"""
    # subs is a list
    assert len(siz) == len(subs), 'found inconsistent siz and subs: %d %d' % (len(siz), len(subs))

    k = np.cumprod(siz[::-1])

    ndx = subs[-1]
    for i, v in enumerate(subs[:-1][::-1]):
        ndx = ndx + v * k[i]

    return ndx



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np
import tensorflow as tf
import keras.layers as KL
from keras.models import Model

# project imports
from lamareg.SynthSeg import evaluate

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import edit_volumes
from lamareg.ext.neuron import models as nrn_models


def predict(
    path_predictions,
    path_qc_results,
    path_model,
    labels_list,
    labels_to_convert=None,
    convert_gt=False,
    shape=224,
    n_levels=5,
    nb_conv_per_level=3,
    conv_size=5,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="relu",
    path_gts=None,
    verbose=True,
):

    # prepare input/output filepaths
    path_predictions, path_gts, path_qc_results, path_gt_results, path_diff = (
        prepare_output_files(path_predictions, path_gts, path_qc_results)
    )

    # get label list
    labels_list, _ = utils.get_list_labels(label_list=labels_list)
    labels_list_unique, _ = np.unique(labels_list, return_index=True)
    if labels_to_convert is not None:
        labels_to_convert, _ = utils.get_list_labels(label_list=labels_to_convert)

    # prepare qc results
    pred_qc_results = np.zeros((len(labels_list_unique) + 1, len(path_predictions)))
    gt_qc_results = (
        np.zeros((len(labels_list_unique), len(path_predictions)))
        if path_gt_results is not None
        else None
    )

    # build network
    model_input_shape = [None, None, None, 1]
    net = build_qc_model(
        path_model=path_model,
        input_shape=model_input_shape,
        label_list=labels_list_unique,
        n_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        unet_feat_count=unet_feat_count,
        feat_multiplier=feat_multiplier,
        activation=activation,
    )

    # perform segmentation
    loop_info = utils.LoopInfo(len(path_predictions), 10, "predicting", True)
    for idx, (path_prediction, path_gt) in enumerate(zip(path_predictions, path_gts)):

        # compute segmentation only if needed
        if verbose:
            loop_info.update(idx)

        # preprocessing
        prediction, gt_scores = preprocess(
            path_prediction, path_gt, shape, labels_list, labels_to_convert, convert_gt
        )

        # get predicted scores
        pred_qc_results[-1, idx] = np.sum(prediction > 0)
        pred_qc_results[:-1, idx] = np.clip(np.squeeze(net.predict(prediction)), 0, 1)
        np.save(path_qc_results, pred_qc_results)

        # save GT scores if necessary
        if gt_scores is not None:
            gt_qc_results[:, idx] = gt_scores
            np.save(path_gt_results, gt_qc_results)

    if path_diff is not None:
        diff = pred_qc_results[:-1, :] - gt_qc_results
        np.save(path_diff, diff)


def prepare_output_files(path_predictions, path_gts, path_qc_results):

    # check inputs
    assert path_predictions is not None, "please specify an input file/folder (--i)"
    assert path_qc_results is not None, "please specify an output file/folder (--o)"

    # convert path to absolute paths
    path_predictions = os.path.abspath(path_predictions)
    path_qc_results = os.path.abspath(path_qc_results)

    # list input predictions
    path_predictions = utils.list_images_in_folder(path_predictions)

    # build path output with qc results
    if path_qc_results[-4:] != ".npy":
        print(
            "Path for QC outputs provided without npy extension. Adding npy extension."
        )
        path_qc_results += ".npy"
    utils.mkdir(os.path.dirname(path_qc_results))

    if path_gts is not None:
        path_gts = utils.list_images_in_folder(path_gts)
        assert len(path_gts) == len(
            path_predictions
        ), "not the same number of predictions and GTs"
        path_gt_results = path_qc_results.replace(".npy", "_gt.npy")
        path_diff = path_qc_results.replace(".npy", "_diff.npy")
    else:
        path_gts = [None] * len(path_predictions)
        path_gt_results = path_diff = None

    return path_predictions, path_gts, path_qc_results, path_gt_results, path_diff


def preprocess(
    path_prediction,
    path_gt=None,
    shape=224,
    labels_list=None,
    labels_to_convert=None,
    convert_gt=False,
):

    # read image and corresponding info
    pred, _, aff_pred, n_dims, _, _, _ = utils.get_volume_info(path_prediction, True)
    gt = utils.load_volume(path_gt, aff_ref=np.eye(4)) if path_gt is not None else None

    # align
    pred = edit_volumes.align_volume_to_ref(
        pred, aff_pred, aff_ref=np.eye(4), n_dims=n_dims
    )

    # pad/crop to 224, such that segmentations are in the middle of the patch
    if gt is not None:
        pred, gt = make_shape(pred, gt, shape, n_dims)
    else:
        pred, _ = edit_volumes.crop_volume_around_region(pred, cropping_shape=shape)

    # convert labels if necessary
    if labels_to_convert is not None:
        lut = utils.get_mapping_lut(labels_to_convert, labels_list)
        pred = lut[pred.astype("int32")]
        if convert_gt & (gt is not None):
            gt = lut[gt.astype("int32")]

    # compute GT dice scores
    gt_scores = (
        evaluate.fast_dice(pred, gt, np.unique(labels_list)) if gt is not None else None
    )

    # add batch and channel axes
    pred = utils.add_axis(
        pred, axis=0
    )  # channel axis will be added later when computing one-hot

    return pred, gt_scores


def make_shape(pred, gt, shape, n_dims):

    mask = ((pred > 0) & (pred != 24)) | (gt > 0)
    vol_shape = np.array(pred.shape[:n_dims])

    if np.any(mask):

        # find cropping indices
        indices = np.nonzero(mask)
        min_idx = np.maximum(np.array([np.min(idx) for idx in indices]), 0)
        max_idx = np.minimum(np.array([np.max(idx) for idx in indices]) + 1, vol_shape)

        # expand/retract (depending on the desired shape) the cropping region around the centre
        intermediate_vol_shape = max_idx - min_idx
        cropping_shape = np.array(utils.reformat_to_list(shape, length=n_dims))
        min_idx = min_idx - np.int32(
            np.ceil((cropping_shape - intermediate_vol_shape) / 2)
        )
        max_idx = max_idx + np.int32(
            np.floor((cropping_shape - intermediate_vol_shape) / 2)
        )

        # crop volume
        cropping = np.concatenate(
            [np.maximum(min_idx, 0), np.minimum(max_idx, vol_shape)]
        )
        pred = edit_volumes.crop_volume_with_idx(pred, cropping, n_dims=n_dims)
        gt = edit_volumes.crop_volume_with_idx(gt, cropping, n_dims=n_dims)

        # check if we need to pad the output to the desired shape
        min_padding = np.abs(np.minimum(min_idx, 0))
        max_padding = np.maximum(max_idx - vol_shape, 0)
        if np.any(min_padding > 0) | np.any(max_padding > 0):
            pad_margins = tuple(
                [(min_padding[i], max_padding[i]) for i in range(n_dims)]
            )
            pred = np.pad(pred, pad_margins, mode="constant", constant_values=0)
            gt = np.pad(gt, pad_margins, mode="constant", constant_values=0)

    return pred, gt


def build_qc_model(
    path_model,
    input_shape,
    label_list,
    n_levels,
    nb_conv_per_level,
    conv_size,
    unet_feat_count,
    feat_multiplier,
    activation,
):

    assert os.path.isfile(path_model), "The provided model path does not exist."
    label_list_unique = np.unique(label_list)
    n_labels = len(label_list_unique)

    # one-hot encoding of the input prediction as the network expects soft probabilities
    input_labels = KL.Input(input_shape[:-1])
    labels = KL.Lambda(
        lambda x: tf.one_hot(tf.cast(x, dtype="int32"), depth=n_labels, axis=-1)
    )(input_labels)
    net = Model(inputs=input_labels, outputs=labels)

    # build model
    model = nrn_models.conv_enc(
        input_model=net,
        input_shape=input_shape,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
        use_residuals=True,
        name="qc",
    )
    last = model.outputs[0]

    conv_kwargs = {
        "padding": "same",
        "activation": "relu",
        "data_format": "channels_last",
    }
    last = KL.MaxPool3D(
        pool_size=(2, 2, 2), name="qc_maxpool_%s" % (n_levels - 1), padding="same"
    )(last)
    last = KL.Conv3D(n_labels, kernel_size=5, **conv_kwargs, name="qc_final_conv_0")(
        last
    )
    last = KL.Conv3D(n_labels, kernel_size=5, **conv_kwargs, name="qc_final_conv_1")(
        last
    )
    last = KL.Lambda(lambda x: tf.reduce_mean(x, axis=[1, 2, 3]), name="qc_final_pred")(
        last
    )

    net = Model(inputs=net.inputs, outputs=last)
    net.load_weights(path_model, by_name=True)

    return net



"""

This code is for training is traditional supervised networks with real images and corresponding ground truth labels.
It's relatively simpler than training.py since it here we do not have to generate synthetic scans. However we kept the
parameters for online augmentation.

If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np
import tensorflow as tf
from keras import models
import keras.layers as KL
import numpy.random as npr

# project imports
from lamareg.SynthSeg import metrics_model as metrics
from lamareg.SynthSeg.training import train_model
from lamareg.SynthSeg.labels_to_image_model import get_shapes

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.neuron import models as nrn_models
from lamareg.ext.lab2im import edit_tensors as l2i_et
from lamareg.ext.lab2im.edit_volumes import get_ras_axes


def training(
    image_dir,
    labels_dir,
    model_dir,
    segmentation_labels=None,
    n_neutral_labels=None,
    subjects_prob=None,
    batchsize=1,
    target_res=None,
    output_shape=None,
    flipping=True,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=4.0,
    nonlin_scale=0.04,
    randomise_res=True,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    data_res=None,
    thickness=None,
    bias_field_std=0.7,
    bias_scale=0.025,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=3,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="elu",
    lr=1e-4,
    wl2_epochs=1,
    dice_epochs=50,
    steps_per_epoch=10000,
    checkpoint=None,
):
    """
    This function trains a UNet to segment MRI images with real scans and corresponding ground truth labels.
    We regroup the parameters in four categories: General, Augmentation, Architecture, Training.

    # IMPORTANT !!!
    # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
    # these values refer to the RAS axes.

    :param image_dir: path of folder with all training images
    :param labels_dir: path of folder with all corresponding label maps
    :param model_dir: path of a directory where the models will be saved during training.

    # ----------------------------------------------- General parameters -----------------------------------------------
    # label maps parameters
    :param n_neutral_labels: (optional) if the label maps contain some right/left specific labels and if flipping is
    applied during training, please provide the number of non-sided labels (including the background).
    This is used to know where the sided labels start in generation_labels. Leave to default (None) if either one of the
    two conditions is not fulfilled.
    :param segmentation_labels: (optional) list of the same length as generation_labels to indicate which values to use
    in the training label maps, i.e. all occurrences of generation_labels[i] in the input label maps will be converted
    to output_labels[i] in the returned label maps. Examples:
    Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
    Set output_labels[i]=generation_labels[i] if you wish to keep the value generation_labels[i] in the returned maps.
    Can be a list or a 1d numpy array, or the path to such an array. Default is output_labels = generation_labels.
    :param subjects_prob: (optional) relative order of importance (doesn't have to be probabilistic), with which to pick
    the provided label maps at each minibatch. Can be a sequence, a 1D numpy array, or the path to such an array, and it
    must be as long as path_label_maps. By default, all label maps are chosen with the same importance.

    # output-related parameters
    :param batchsize: (optional) number of images to generate per mini-batch. Default is 1.
    :param target_res: (optional) target resolution at which to teach the network to segment.
    If None, this will be the resolution of the given images/label maps.
    Can be a number (isotropic resolution), or the path to a 1d numpy array.
    :param output_shape: (optional) desired shape of the output image, obtained by randomly cropping the generated image
    Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
    Default is None, where no cropping is performed.

    # --------------------------------------------- Augmentation parameters --------------------------------------------
    # spatial deformation parameters
    :param flipping: (optional) whether to introduce right/left random flipping. Default is True.
    :param scaling_bounds: (optional) if apply_linear_trans is True, the scaling factor for each dimension is
    sampled from a uniform distribution of predefined bounds. Can either be:
    1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
    (1-scaling_bounds, 1+scaling_bounds) for each dimension.
    2) the path to a numpy array of shape (2, n_dims), in which case the scaling factor in dimension i is sampled from
    the uniform distribution of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
    3) False, in which case scaling is completely turned off.
    Default is scaling_bounds = 0.2 (case 1)
    :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for case 1 the
    bounds are centred on 0 rather than 1, i.e. (0+rotation_bounds[i], 0-rotation_bounds[i]).
    Default is rotation_bounds = 15.
    :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
    :param translation_bounds: (optional) same as scaling bounds. Default is translation_bounds = False, but we
    encourage using it when cropping is deactivated (i.e. when output_shape=None).
    :param nonlin_std: (optional) Standard deviation of the normal distribution from which we sample the first
    tensor for synthesising the deformation field. Set to 0 to completely deactivate elastic deformation.
    :param nonlin_scale: (optional) Ratio between the size of the input label maps and the size of the sampled
    tensor for synthesising the elastic deformation field.

    # blurring/resampling parameters
    :param randomise_res: (optional) whether to mimic images that would have been 1) acquired at low resolution, and
    2) resampled to high resolution. The low resolution is uniformly resampled at each minibatch from [1mm, 9mm].
    In that process, the images generated by sampling the GMM are: 1) blurred at the sampled LR, 2) downsampled at LR,
    and 3) resampled at target_resolution.
    :param max_res_iso: (optional) If randomise_res is True, this enables to control the upper bound of the uniform
    distribution from which we sample the random resolution U(min_res, max_res_iso), where min_res is the resolution of
    the input label maps. Must be a number, and default is 4. Set to None to deactivate it, but if randomise_res is
    True, at least one of max_res_iso or max_res_aniso must be given.
    :param max_res_aniso: If randomise_res is True, this enables to downsample the input volumes to a random LR in
    only 1 (random) direction. This is done by randomly selecting a direction i in the range [0, n_dims-1], and sampling
    a value in the corresponding uniform distribution U(min_res[i], max_res_aniso[i]), where min_res is the resolution
    of the input label maps. Can be a number, a sequence, or a 1d numpy array. Set to None to deactivate it, but if
    randomise_res is True, at least one of max_res_iso or max_res_aniso must be given.
    :param data_res: (optional) specific acquisition resolution to mimic, as opposed to random resolution sampled when
    randomise_res is True. This triggers a blurring which mimics the acquisition resolution, but downsampling is
    optional (see param downsample). Default for data_res is None, where images are slightly blurred. If the generated
    images are uni-modal, data_res can be a number (isotropic acquisition resolution), a sequence, a 1d numpy array, or
    the path to a 1d numpy array. In the multi-modal case, it should be given as a numpy array (or a path) of size
    (n_mod, n_dims), where each row is the acquisition resolution of the corresponding channel.
    :param thickness: (optional) if data_res is provided, we can further specify the slice thickness of the low
    resolution images to mimic. Must be provided in the same format as data_res. Default thickness = data_res.

    # bias field parameters
    :param bias_field_std: (optional) If strictly positive, this triggers the corruption of images with a bias field.
    The bias field is obtained by sampling a first small tensor from a normal distribution, resizing it to
    full size, and rescaling it to positive values by taking the voxel-wise exponential. bias_field_std designates the
    std dev of the normal distribution from which we sample the first tensor.
    Set to 0 to completely deactivate bias field corruption.
    :param bias_scale: (optional) If bias_field_std is not False, this designates the ratio between the size of
    the input label maps and the size of the first sampled tensor for synthesising the bias field.

    # ------------------------------------------ UNet architecture parameters ------------------------------------------
    :param n_levels: (optional) number of level for the Unet. Default is 5.
    :param nb_conv_per_level: (optional) number of convolutional layers per level. Default is 2.
    :param conv_size: (optional) size of the convolution kernels. Default is 2.
    :param unet_feat_count: (optional) number of feature for the first layer of the UNet. Default is 24.
    :param feat_multiplier: (optional) multiply the number of feature by this number at each new level. Default is 2.
    :param activation: (optional) activation function. Can be 'elu', 'relu'.

    # ----------------------------------------------- Training parameters ----------------------------------------------
    :param lr: (optional) learning rate for the training. Default is 1e-4
    :param wl2_epochs: (optional) number of epochs for which the network (except the soft-max layer) is trained with L2
    norm loss function. Default is 1.
    :param dice_epochs: (optional) number of epochs with the soft Dice loss function. Default is 50.
    :param steps_per_epoch: (optional) number of steps per epoch. Default is 10000. Since no online validation is
    possible, this is equivalent to the frequency at which the models are saved.
    :param checkpoint: (optional) path of an already saved model to load before starting the training.
    """

    # check epochs
    assert (wl2_epochs > 0) | (
        dice_epochs > 0
    ), "either wl2_epochs or dice_epochs must be positive, had {0} and {1}".format(
        wl2_epochs, dice_epochs
    )

    # prepare data files
    path_images = utils.list_images_in_folder(image_dir)
    path_labels = utils.list_images_in_folder(labels_dir)
    assert len(path_images) == len(
        path_labels
    ), "There should be as many images as label maps."

    # get label lists
    label_list, _ = utils.get_list_labels(
        label_list=segmentation_labels, labels_dir=labels_dir
    )
    n_labels = np.size(label_list)

    # create augmentation model
    im_shape, _, _, n_channels, _, atlas_res = utils.get_volume_info(
        path_images[0], aff_ref=np.eye(4)
    )
    augmentation_model = build_augmentation_model(
        im_shape,
        n_channels,
        label_list,
        n_neutral_labels,
        atlas_res,
        target_res,
        output_shape=output_shape,
        output_div_by_n=2**n_levels,
        flipping=flipping,
        aff=np.eye(4),
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        randomise_res=randomise_res,
        max_res_iso=max_res_iso,
        max_res_aniso=max_res_aniso,
        data_res=data_res,
        thickness=thickness,
        bias_field_std=bias_field_std,
        bias_scale=bias_scale,
    )
    unet_input_shape = augmentation_model.output[0].get_shape().as_list()[1:]

    # prepare the segmentation model
    unet_model = nrn_models.unet(
        input_model=augmentation_model,
        input_shape=unet_input_shape,
        nb_labels=n_labels,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
        name="unet",
    )

    # input generator
    generator = build_model_inputs(path_images, path_labels, batchsize, subjects_prob)
    input_generator = utils.build_training_generator(generator, batchsize)

    # pre-training with weighted L2, input is fit to the softmax rather than the probabilities
    if wl2_epochs > 0:
        wl2_model = models.Model(
            unet_model.inputs, [unet_model.get_layer("unet_likelihood").output]
        )
        wl2_model = metrics.metrics_model(wl2_model, label_list, "wl2")
        train_model(
            wl2_model,
            input_generator,
            lr,
            wl2_epochs,
            steps_per_epoch,
            model_dir,
            "wl2",
            checkpoint,
        )
        checkpoint = os.path.join(model_dir, "wl2_%03d.h5" % wl2_epochs)

    # fine-tuning with dice metric
    dice_model = metrics.metrics_model(unet_model, label_list, "dice")
    train_model(
        dice_model,
        input_generator,
        lr,
        dice_epochs,
        steps_per_epoch,
        model_dir,
        "dice",
        checkpoint,
    )


def build_augmentation_model(
    im_shape,
    n_channels,
    segmentation_labels,
    n_neutral_labels,
    atlas_res,
    target_res,
    output_shape=None,
    output_div_by_n=None,
    flipping=True,
    aff=None,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=4.0,
    nonlin_scale=0.0625,
    randomise_res=False,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    data_res=None,
    thickness=None,
    bias_field_std=0.7,
    bias_scale=0.025,
):

    # reformat resolutions and get shapes
    im_shape = utils.reformat_to_list(im_shape)
    n_dims, _ = utils.get_dims(im_shape)
    if data_res is not None:
        data_res = utils.reformat_to_n_channels_array(data_res, n_dims, n_channels)
        thickness = (
            data_res
            if thickness is None
            else utils.reformat_to_n_channels_array(thickness, n_dims, n_channels)
        )
        target_res = (
            atlas_res
            if (target_res is None)
            else utils.reformat_to_n_channels_array(target_res, n_dims)[0]
        )
    else:
        target_res = atlas_res

    # get shapes
    crop_shape, output_shape = get_shapes(
        im_shape, output_shape, atlas_res, target_res, output_div_by_n
    )

    # define model inputs
    image_input = KL.Input(shape=im_shape + [n_channels], name="image_input")
    labels_input = KL.Input(shape=im_shape + [1], name="labels_input", dtype="int32")

    # deform labels
    labels, image = layers.RandomSpatialDeformation(
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        inter_method=["nearest", "linear"],
    )([labels_input, image_input])

    # cropping
    if crop_shape != im_shape:
        labels, image = layers.RandomCrop(crop_shape)([labels, image])

    # flipping
    if flipping:
        assert aff is not None, "aff should not be None if flipping is True"
        labels, image = layers.RandomFlip(
            get_ras_axes(aff, n_dims)[0],
            [True, False],
            segmentation_labels,
            n_neutral_labels,
        )([labels, image])

    # apply bias field
    if bias_field_std > 0:
        image = layers.BiasFieldCorruption(bias_field_std, bias_scale, False)(image)

    # intensity augmentation
    image = layers.IntensityAugmentation(
        6, clip=False, normalise=True, gamma_std=0.5, separate_channels=True
    )(image)

    # if necessary, loop over channels to 1) blur, 2) downsample to simulated LR, and 3) upsample to target
    if data_res is not None:
        channels = list()
        split = (
            KL.Lambda(lambda x: tf.split(x, [1] * n_channels, axis=-1))(image)
            if (n_channels > 1)
            else [image]
        )
        for i, channel in enumerate(split):

            if randomise_res:
                max_res_iso = np.array(
                    utils.reformat_to_list(max_res_iso, length=n_dims, dtype="float")
                )
                max_res_aniso = np.array(
                    utils.reformat_to_list(max_res_aniso, length=n_dims, dtype="float")
                )
                max_res = np.maximum(max_res_iso, max_res_aniso)
                resolution, blur_res = layers.SampleResolution(
                    atlas_res, max_res_iso, max_res_aniso
                )(image)
                sigma = l2i_et.blurring_sigma_for_downsampling(
                    atlas_res, resolution, thickness=blur_res
                )
                channel = layers.DynamicGaussianBlur(
                    0.75 * max_res / np.array(atlas_res), 1.03
                )([channel, sigma])
                channel = layers.MimicAcquisition(
                    atlas_res, atlas_res, output_shape, False
                )([channel, resolution])
                channels.append(channel)

            else:
                sigma = l2i_et.blurring_sigma_for_downsampling(
                    atlas_res, data_res[i], thickness=thickness[i]
                )
                channel = layers.GaussianBlur(sigma, 1.03)(channel)
                resolution = KL.Lambda(
                    lambda x: tf.convert_to_tensor(data_res[i], dtype="float32")
                )([])
                channel = layers.MimicAcquisition(atlas_res, data_res[i], output_shape)(
                    [channel, resolution]
                )
                channels.append(channel)

        # concatenate all channels back
        image = (
            KL.Lambda(lambda x: tf.concat(x, -1))(channels)
            if len(channels) > 1
            else channels[0]
        )

        # resample labels at target resolution
        if crop_shape != output_shape:
            labels = l2i_et.resample_tensor(
                labels, output_shape, interp_method="nearest"
            )

    # build model (dummy layer enables to keep the labels when plugging this model to other models)
    labels = KL.Lambda(lambda x: tf.cast(x, dtype="int32"), name="labels_out")(labels)
    image = KL.Lambda(lambda x: x[0], name="image_out")([image, labels])
    brain_model = models.Model(
        inputs=[image_input, labels_input], outputs=[image, labels]
    )

    return brain_model


def build_model_inputs(
    path_inputs,
    path_outputs,
    batchsize=1,
    subjects_prob=None,
    dtype_input="float32",
    dtype_output="int32",
):

    # get label info
    _, _, _, n_channels, _, _ = utils.get_volume_info(path_inputs[0])

    # make sure subjects_prob sums to 1
    if subjects_prob is not None:
        subjects_prob /= np.sum(subjects_prob)

    # Generate!
    while True:

        # randomly pick as many images as batchsize
        indices = npr.choice(
            np.arange(len(path_outputs)), size=batchsize, p=subjects_prob
        )

        # initialise input lists
        list_batch_inputs = list()
        list_batch_outputs = list()

        for idx in indices:

            # get a batch input
            batch_input = utils.load_volume(
                path_inputs[idx], aff_ref=np.eye(4), dtype=dtype_input
            )
            if n_channels > 1:
                list_batch_inputs.append(utils.add_axis(batch_input, axis=0))
            else:
                list_batch_inputs.append(utils.add_axis(batch_input, axis=[0, -1]))

            # get a batch output
            batch_output = utils.load_volume(
                path_outputs[idx], aff_ref=np.eye(4), dtype=dtype_output
            )
            list_batch_outputs.append(utils.add_axis(batch_output, axis=[0, -1]))

        # build list of training pairs
        list_training_pairs = [list_batch_inputs, list_batch_outputs]
        if batchsize > 1:  # concatenate individual input types if batchsize > 1
            list_training_pairs = [
                np.concatenate(item, 0) for item in list_training_pairs
            ]
        else:
            list_training_pairs = [item[0] for item in list_training_pairs]

        yield list_training_pairs
